# -*- coding: utf-8 -*-
import unittest
import circadapt.error
from circadapt import CircAdapt
import numpy as np
import matplotlib.pyplot as plt
import pytest

from _settings import test_dts, test_solver, settings

# %% analytical solution
def analytical_solution(t_true, model):
    """
    Analytical solution of ArtVen.

    Parameters
    ----------
    t_true : array
        timesteps to caluclate solution.
    model : circadapt.model.benchmark.ArtVen
        Model to get parameters from.

    Returns
    -------
    array
        analytical solution of flow.

    """
    VProx_true = t_true*0

    a=model.get('Model.NodeProx.l')

    p0av = model.get('Model.V.p0')
    q0av = model.get('Model.V.q0')
    p0 = model.get('Model.NodeProx.p0')
    A0 = model.get('Model.NodeProx.A0')
    AWall = model.get('Model.NodeProx.A_wall')
    Len = model.get('Model.NodeProx.l')

    V = model['Cavity']['V']
    if len(V.shape)==2:
        V = V[0, :]
    VProx0 = V[0]
    VDist0 = V[1]
    Vtot = VProx0+VDist0


    par_k = model.get('Model.NodeProx.k')

    l = p0av / p0 * (A0/AWall+0.5) / q0av * AWall



    if par_k==1:
        VProx_true = np.exp((-2*t_true)/(a*l))
        VProx_true = VProx_true / VProx_true[0] * (VProx0 - (VProx0+VDist0)/2 )
        VProx_true = VProx_true + (Vtot)/2
    elif par_k==2:
        VProx_true = np.exp( (2 * Vtot * t_true) / (a**2 * l) + (-2*t_true)/(a*l))
        VProx_true = VProx_true / VProx_true[0] * (VProx0 - (VProx0+VDist0)/2 )
        VProx_true = VProx_true + (Vtot)/2

    # VDist is per definition Vtot - VProx
    VDist_true = Vtot-VProx_true

    # Pressure in both cavities
    # Vcor = AWall/VProx_true/Len
    # Vcor[Vcor<0]=0
    # pProx_true = p0*(((VProx_true+0.5*AWall*Len)/(A0*Len+0.5*AWall*Len))**par_k - Vcor**2 )

    # Vcor = AWall/VDist_true/Len
    # Vcor[Vcor<0]=0
    # pDist_true = p0*(((VDist_true+0.5*AWall*Len)/(A0*Len+0.5*AWall*Len))**par_k - Vcor**2 )

    pProx_true = p0*((VProx_true/Len/AWall+0.5) / (A0/AWall+0.5))**par_k
    pDist_true = p0*((VDist_true/Len/AWall+0.5) / (A0/AWall+0.5))**par_k


    q_true = q0av * ((pProx_true - pDist_true) / p0av)**par_k


    # q_true = t_true * 0
    return {'t': t_true, 'q': q_true*1e3,
            'V_prox': VProx_true*1e6, 'V_dist': VDist_true*1e6}


# %% simulate
def create_model(DT, solver):
    order = None
    if solver[:9] == "SolverBDF":
        order = int(solver[9:])
        solver = "SolverBDF"

    solver_name = solver[:]
    if solver_name[-1].isnumeric() and solver_name[-3:-1] == '_o':
        order = int(solver_name[-1])
        solver_name = solver_name[:-3]

    model = CircAdapt(solver_name)

    if order is not None:
        model.set('Solver.order', order)

    model.set('Solver.dt', DT)
    model.set('Solver.dt_export', DT)

    model.add_component('Tube0D', 'NodeProx')
    model.add_component('Tube0D', 'NodeDist')
    model.add_component('ArtVen', 'V')

    model.set_component('V.prox', 'NodeProx')
    model.set_component('V.dist', 'NodeDist')

    VProx0 = 350e-6
    VDist0 = 300e-6

    par_k = 1

    set_pars = [
            ('Model.t_cycle', 1),
            ('Model.NodeProx.V', VProx0),
            ('Model.NodeDist.V', VDist0),
            ('Model.NodeProx.k', par_k),
            ('Model.NodeDist.k', par_k),
            ('Model.V.p0', 1000),
            ('Model.V.k', 1.),
            ]
    for set_par in set_pars:
        model.set(*set_par)
    return model


def get_model_signals(model):
    signals = {'t': model.get('Solver.t')}
    signals['q'] = model['ArtVen']['q'][:, 0] * 1e3
    signals['V_prox'] = model['Cavity']['V'][:, 0] * 1e6
    signals['V_dist'] = model['Cavity']['V'][:, 1] * 1e6


    return signals


def calculate_errors(dt, solver, signal):
    try:
        model = create_model(dt, solver)
        model.run(1)
        t_true = model.get('Solver.t')
        true_signals = analytical_solution(t_true, model)
        model_signals = get_model_signals(model)
        return np.sqrt(np.mean(
            ((true_signals[signal] - model_signals[signal]) /
             np.max(np.abs(true_signals[signal]))
             )**2))
    except circadapt.error.ModelCrashed:
        return np.nan

    
# %% Pytest
@pytest.mark.parametrize("dt", test_dts, ids=lambda d: f"dt={d*1e3}ms")
@pytest.mark.parametrize("solver", test_solver, ids=lambda d: f"solver={d}")
def test_artven(dt, solver):
    """
    Test elements:
        Build Pre-Afterload module
        Build Patch2022
        Set parameters of Patch2022
        Run 1 beat
    """
    # which signal to compare
    signal_to_test = 'q'

    # run tests
    e = 1e-3
    assert np.abs(calculate_errors(dt, solver, signal_to_test)) < e


# %% main
if __name__ == "__main__":
    # plot settings
    plot_dt = np.array([1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1])
    plot_solvers = test_solver


    # get options
    model = create_model(1e-3, 'forward_euler')

    # get all errors
    error = np.ndarray((len(plot_solvers), len(plot_dt)))
    errorV = np.ndarray((len(plot_solvers), len(plot_dt)))
    for i_solver, solver in enumerate(plot_solvers):
        for i_dt, dt in enumerate(plot_dt):
            print(f'\rsimulate data solver {i_solver} with Delta t = {dt}    ',
                  end='\r')
            error[i_solver, i_dt] = calculate_errors(dt, solver, 'q')
            errorV[i_solver, i_dt] = calculate_errors(dt, solver, 'V_prox')
    print('Simulated all data                                                ')
    print(error)

    # %% make plot
    t_true = np.linspace(0, model.get('Model.t_cycle'), 10000)
    model = create_model(1e-4, plot_solvers[0])
    model.run(1)
    true_signals = analytical_solution(t_true, model)
    model_signals = get_model_signals(model)

    # Create a new figure
    plt.figure(1, clear=True, figsize=(8, 8))

    # Plot signal
    ax1 = plt.subplot(2, 2, 1)
    # plt.plot(true_signals['t'],
    #          true_signals['q'],
    #          label='True Signals', ls='--')
    plt.plot(model_signals['t']*1e3,
             model_signals['q'],
             label='Model Signals', ls='-', lw=1)
    plt.title('Flow', fontsize=14)
    plt.ylabel('$q [L/s]$', fontsize=12)
    plt.xlabel('$t [ms]$', fontsize=12)


    ax2 = plt.subplot(2, 2, 2)
    # plt.plot(true_signals['t'],
    #          true_signals['V_prox'],
    #          label='True Signals', ls='--', c='r')
    # plt.plot(true_signals['t'],
    #          true_signals['V_dist'],
    #          label='True Signals', ls='--', c='b')
    plt.plot(model_signals['t']*1e3,
             model_signals['V_prox'],
             label='$V_{prox}$', ls='-', lw=1, c='r')
    plt.plot(model_signals['t']*1e3,
             model_signals['V_dist'],
             label='$V_{dist}$', ls='-', lw=1, c='b')
    plt.title('Proximal (red) and Distal (blue) volume ', fontsize=14)
    plt.ylabel('$V [mL]$', fontsize=12)
    plt.xlabel('$t [ms]$', fontsize=12)
    plt.legend()

    # Plot L2
    ax3 = plt.subplot(2, 2, 3)
    for i in range(error.shape[0]):
        ax3.plot(plot_dt*1e3, error[i, :], c=settings['solver_color'][plot_solvers[i]])
    plt.yscale('log')
    plt.xscale('log')

    # plt.axhline(1e-2, c='k', ls='--')
    # plt.axvline(1e-3, c='k', ls='--')
    plt.title('Mean absolute error', fontsize=14)
    plt.xlabel('${\\Delta}t [ms]$', fontsize=12)
    plt.ylabel('Mean absolute error [-]')


    ax4 = plt.subplot(2, 2, 4)
    for i in range(errorV.shape[0]):
        ax4.plot(plot_dt*1e3, errorV[i, :], c=settings['solver_color'][plot_solvers[i]])
    plt.yscale('log')
    plt.xscale('log')

    # plt.axhline(1e-2, c='k', ls='--')
    # plt.axvline(1e-3, c='k', ls='--')


    # Move the legend outside the plot to the right
    # plt.legend(plot_solvers, bbox_to_anchor=(1.05, 1), loc='upper left')
    plt.legend(plot_solvers, loc='lower left', bbox_to_anchor=(0.35, 0.0, 0.5, 0.5))

    plt.title('Mean absolute error', fontsize=14)
    plt.xlabel('${\\Delta}t [ms]$', fontsize=12)
    plt.ylabel('Mean absolute error [-]')

    # Super title for the entire figure
    plt.suptitle('Verification of ArtVen', fontsize=18, weight='bold')

    # Remove right and top spines of the first subplot
    for ax in [ax1, ax2, ax3, ax4]:
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)


    # Adjust layout
    # plt.tight_layout()
    plt.subplots_adjust(top=0.92,
        bottom=0.079,
        left=0.15,
        right=0.9,
        hspace=0.285,
        wspace=0.2)


    # Show the plot
    plt.show()
